#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
The device class `Technix` and its corresponding configuration class
"""

import logging
from time import sleep
from typing import cast

from hvl_ccb import configdataclass
from hvl_ccb.dev import SingleCommDevice
from hvl_ccb.utils.poller import Poller
from hvl_ccb.utils.typing import Number
from hvl_ccb.utils.validation import validate_bool, validate_number

from .base import (
    TechnixError,
    TechnixFaultError,
    _GetRegisters,
    _SetRegisters,
    _Status,
    _TechnixCommunicationClasses,
)

logger = logging.getLogger(__name__)


@configdataclass
class TechnixConfig:
    #: communication channel between computer and Technix
    communication_channel: _TechnixCommunicationClasses

    #: Maximal Output voltage
    max_voltage: Number

    #: Maximal Output current
    max_current: Number

    #: Polling interval in s to maintain to watchdog of the device
    polling_interval_sec: Number = 4

    #: Time to wait after stopping the device
    post_stop_pause_sec: Number = 1

    #: Time for pulsing a register
    register_pulse_time: Number = 0.1

    #: Read output voltage and current within the polling event
    read_output_while_polling: bool = False


class Technix(SingleCommDevice):
    """
    Device class to control capacitor chargers from Technix
    """

    def __init__(self, com, dev_config) -> None:
        # Call superclass constructor
        super().__init__(com, dev_config)

        # maximum output current of the hardware
        self._max_current_hardware = self.config.max_current
        # maximum output voltage of the hardware
        self._max_voltage_hardware = self.config.max_voltage

        self._set_voltage: Number = 0
        self._set_current: Number = 0

        #: status of Technix
        self._status: _Status | None = None

        #: Status Poller to maintain the watchdog of the device
        self._status_poller: Poller = Poller(
            spoll_handler=self._spoll_handler,
            polling_interval_sec=self.config.polling_interval_sec,
        )

        logger.debug("Technix Power Supply initialised.")

    @staticmethod
    def config_cls():
        return TechnixConfig

    def default_com_cls(self) -> _TechnixCommunicationClasses:  # type: ignore[override]
        return self.config.communication_channel

    @property
    def is_started(self) -> bool:
        """
        Is the device started?
        """
        return self.status is not None

    def start(self) -> None:
        """
        Start the device and set it into the remote controllable mode. The high
        voltage is turn off, and the status poller is started.
        """

        if self.is_started:
            logger.debug("Technix device was already started.")
            return

        super().start()

        with self.com.access_lock:
            logger.debug("Starting Technix...")
            self.remote = True
            self.output = False
            self.inhibit = False
            self._status_poller.start_polling()

        logger.info("Started Technix")

    def stop(self) -> None:
        """
        Stop the device. The status poller is stopped and the high voltage output
        is turn off.
        """
        if not self.is_started:
            logger.debug("Technix device was not started.")
            return

        with self.com.access_lock:
            self._status_poller.stop_polling()
            self.output = False
            self.remote = False
            self._status = None
            sleep(self.config.post_stop_pause_sec)

        super().stop()

        logger.info("Stopped Technix")

    def _set_register(self, register: _SetRegisters, value: bool | int):
        """
        Function to set a value to a register
        """
        command = f"{register},{int(value)}"
        answer = self.com.query(command)
        if answer != command:
            msg = f"Expected '{command}', but answer was '{answer}'"
            logger.error(msg)
            raise TechnixError(msg)

    def _get_register(self, register: _GetRegisters) -> int:
        """
        Function to query a register
        """
        _register: str = str(register)
        answer = self.com.query(_register)
        if answer[: len(_register)] != _register:
            msg = f"Expected '{_register}', but answer was '{answer}'"
            logger.error(msg)
            raise TechnixError(msg)

        return int(answer[len(_register) :])

    def _spoll_handler(self):
        """
        Function to be called from the poller
        """
        with self.com.access_lock:
            """
            This method can be called manually and by the poller. In case of a
            fault the communication is stopped and subsequent calls of this function
            are skipped. Only one unique call of this function is allowed to be
            performed at the same time. cf. issue 161 on gitlab.com/ethz_hvl/hvl_ccb
            """
            try:
                self.query_status()
                logger.info(self.status)
            except TechnixError as e:
                self._status_poller.stop_polling()
                self._status = None
                msg = (
                    "An error occurred during the polling event and the connection "
                    "is closed"
                )
                logger.exception(msg)
                raise TechnixError(msg) from e

    @property
    def status(self) -> _Status | None:
        """
        The status of the device with the different states as sub-fields
        """
        return self._status if self.com.is_open else None

    @property
    def max_current(self) -> Number:
        """
        Maximal output current of the hardware in A
        """
        return self._max_current_hardware

    @property
    def max_voltage(self) -> Number:
        """
        Maximal output voltage of the hardware in V
        """
        return self._max_voltage_hardware

    def query_status(self, *, _retry: bool = False) -> None:
        """
        Query the status of the device.

        :return: This function returns nothing
        """
        with self.com.access_lock:
            """
            This method can be called manually and by the poller. In case of a
            fault the communication is stopped and subsequent calls of this function
            are skipped. Only one unique call of this function is allowed to be
            performed at the same time. cf. issue 161 on gitlab.com/ethz_hvl/hvl_ccb
            """
            try:
                status_byte = self._get_register(_GetRegisters.STATUS)  # type: ignore[arg-type]
                validate_number(
                    "Integer of status byte", status_byte, (0, 255), int, logger
                )
                status_bits = [bool(int(bit)) for bit in f"{status_byte:08b}"]
                status_bits[1] = not status_bits[1]

                if self.config.read_output_while_polling:
                    voltage: Number | None = self._measure_voltage()
                    current: Number | None = self._measure_current()
                else:
                    voltage, current = None, None

                self._status = _Status(*status_bits, voltage, current)  # type: ignore[call-arg]

                if self._status.fault and not self.open_interlock:
                    msg = (
                        "The fault flag was detected with closed interlock. There can"
                        "be a hardware problem with the device."
                    )
                    logger.error(msg)
                    raise TechnixFaultError(msg)  # noqa: TRY301

            except TechnixFaultError as e:
                if not _retry:
                    # When the interlock gets closed, the HV has to be turned off to
                    # remove the fault flag. This is achieved with this retry.
                    logger.info(
                        "Try to clear the fault. If it persists, there is a "
                        "real hardware problem."
                    )
                    self.output = False
                    self.query_status(_retry=True)
                else:
                    self._status = None
                    msg = "An error occurred during querying the status"
                    logger.exception(msg)
                    raise TechnixError(msg) from e

    def _measure_voltage(self) -> Number:
        """
        Internal function to measure the output voltage
        """
        return self._get_register(_GetRegisters.VOLTAGE) / 4095 * self.max_voltage  # type: ignore[arg-type]

    @property
    def voltage(self) -> Number:
        """
        Actual voltage at the output in V
        """
        if not self.is_started:
            return 0
        if self.config.read_output_while_polling:
            _voltage = cast("Number", cast("_Status", self.status).voltage)
            # Make mypy happy, `_voltage` is None when it is not queried during polling
        else:
            _voltage = self._measure_voltage()
        logger.info(f"Present Output Voltage: {_voltage:_.2f} V")
        return _voltage

    @voltage.setter
    def voltage(self, value: Number):
        """
        Set voltage of the high voltage output

        :param value: Voltage as a `Number` in V
        :raises ValueError: if the set voltage is below 0 V or higher than the
            maximal voltage of the device
        """
        validate_number("Set Voltage", value, (0, self.max_voltage), logger=logger)
        _voltage = int(4095 * value / self.max_voltage)
        validate_number(
            "Register value of set voltage", _voltage, (0, 4095), int, logger=logger
        )
        # Double-check the value if it is really within the limits

        self._set_voltage = value
        logger.info(f"Set Output Voltage: {value:_.2f} V")
        self._set_register(_SetRegisters.VOLTAGE, _voltage)  # type: ignore[arg-type]

    @property
    def set_voltage(self) -> Number:
        """Return the set voltage (may differ from actual value) in V"""
        return self._set_voltage

    @set_voltage.setter
    def set_voltage(self, value: Number) -> None:
        """Set the output voltage"""
        self.voltage = value

    def _measure_current(self) -> Number:
        """
        Internal function to measure the output current
        """
        return self._get_register(_GetRegisters.CURRENT) / 4095 * self.max_current  # type: ignore[arg-type]

    @property
    def current(self) -> Number:
        """
        Actual current of the output in A
        """
        if not self.is_started:
            return 0
        if self.config.read_output_while_polling:
            _current = cast("Number", cast("_Status", self.status).current)
            # Make mypy happy, `_current` is None when it is not queried during polling
        else:
            _current = self._measure_current()
        logger.info(f"Present Output Current: {_current:_.3f} A")
        return _current

    @current.setter
    def current(self, value: Number):
        """
        Set current of the output

        :param value: Current as a `Number` in A
        """
        validate_number("Set Current", value, (0, self.max_current), logger=logger)
        _current = int(4095 * value / self.max_current)
        validate_number(
            "Register value of set current", _current, (0, 4095), int, logger=logger
        )
        # Double-check the value if it is really within the limits

        self._set_current = value
        logger.info(f"Set Output Current: {value:_.3f} A")
        self._set_register(_SetRegisters.CURRENT, _current)  # type: ignore[arg-type]

    @property
    def set_current(self) -> Number:
        """Return the set current (may differ from actual value) in A"""
        return self._set_current

    @set_current.setter
    def set_current(self, value: Number) -> None:
        """Set the output current"""
        self.current = value

    @property
    def output(self) -> bool | None:
        """
        State of the high voltage output
        """
        if self.is_started:
            return cast("_Status", self.status).output
        return None

    @output.setter
    def output(self, value: bool):
        """
        Activates the output of the source

        :param value: `True` for activation, `False` for deactivation
        :raises TypeError: if value is not a `bool`
        """
        validate_bool("Enable HV-Output", value, logger)
        register = _SetRegisters.HVON if value else _SetRegisters.HVOFF

        self._set_register(register, True)  # type: ignore[arg-type]
        sleep(self.config.register_pulse_time)
        self._set_register(register, False)  # type: ignore[arg-type]

        logger.info(f"HV-Output is {'' if value else 'de'}activated")

    @property
    def remote(self) -> bool | None:
        """
        Is the device in remote control mode?
        """
        if self.is_started:
            return cast("_Status", self.status).remote
        return None

    @remote.setter
    def remote(self, value: bool):
        """
        (De-)Activate the remote control mode

        :param value: `True` to control the device with this remote control, `False` to
            control it with the hardware front panel
        """
        validate_bool("Remote control", value, logger)

        self._set_register(_SetRegisters.LOCAL, not value)  # type: ignore[arg-type]
        logger.info(f"Remote control is {'' if value else 'de'}activated")

    @property
    def inhibit(self) -> bool | None:
        """
        Is the output of the voltage inhibited?
        The output stage can still be active.
        """
        if self.is_started:
            return cast("_Status", self.status).inhibit
        return None

    @inhibit.setter
    def inhibit(self, value: bool):
        """
        Inhibit the output without deactivating the HV-output section. To generate
        high voltage this value must be `False`(!).

        :param value: `True` to turn off the output for a short time, `False` to re-turn
            it on
        :raises TypeError: if value is not a `bool`
        """
        validate_bool("Inhibit the output", value, logger)

        self._set_register(_SetRegisters.INHIBIT, value)  # type: ignore[arg-type]
        logger.info(f"Inhibit is {'' if value else 'de'}activated")

    @property
    def open_interlock(self) -> bool | None:
        """
        Is the interlock open? (in safe mode)
        """
        if self.is_started:
            return cast("_Status", self.status).open_interlock
        return None

    @property
    def voltage_regulation(self) -> bool | None:
        """
        Status if the output is in voltage regulation mode (or current regulation)
        """
        if self.is_started:
            return cast("_Status", self.status).voltage_regulation
        return None



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
import socket


def get_free_tcp_port(host):
    with socket.socket() as sock:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((host, 0))
        _addr, port = sock.getsockname()
    return port



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Masked version of SerialCommunication for testing purposes.
"""

from queue import Queue

from hvl_ccb.comm.serial import SerialCommunication
from hvl_ccb.dev.crylas import CryLasLaserSerialCommunication
from hvl_ccb.dev.fug.comm import FuGSerialCommunication
from hvl_ccb.dev.heinzinger import HeinzingerSerialCommunication
from hvl_ccb.dev.ka3000p.comm import KA3000PCommunication
from hvl_ccb.dev.newport import NewportSMC100PPSerialCommunication
from hvl_ccb.dev.pfa2_filter import Pfa2FilterSerialCommunication
from hvl_ccb.dev.pfeiffer_tpg import PfeifferTPGSerialCommunication
from hvl_ccb.dev.sst_luminox import LuminoxSerialCommunication
from hvl_ccb.dev.technix import TechnixSerialCommunication


class LoopSerialCommunication(SerialCommunication):
    """
    Serial communication for the tests with "loop://" port. Masks `write` method
    and adds `put_text` method to put actual values for the serial communication
    protocol to read with the `read_text` method.
    """

    def __init__(self, configuration) -> None:
        super().__init__(configuration)

        self._write_buffer = Queue()

    def write(
        self,
        text: str,
    ) -> None:
        self._write_buffer.put(text)

    def put_text(self, text: str) -> None:
        # super().write(text)
        super().write_bytes(
            text.encode(encoding=self.config.encoding) + self.config.terminator
        )

    def write_bytes(self, data: bytes) -> None:
        self._write_buffer.put(data)

    def put_bytes(self, data: bytes) -> None:
        super().write_bytes(data)

    def get_written(self):
        return self._write_buffer.get() if not self._write_buffer.empty() else None


class CryLasLaserLoopSerialCommunication(
    CryLasLaserSerialCommunication, LoopSerialCommunication
):
    pass


class FuGLoopSerialCommunication(FuGSerialCommunication, LoopSerialCommunication):
    pass


class HeinzingerLoopSerialCommunication(
    HeinzingerSerialCommunication, LoopSerialCommunication
):
    pass


class LuminoxLoopSerialCommunication(
    LuminoxSerialCommunication, LoopSerialCommunication
):
    pass


class NewportLoopSerialCommunication(
    NewportSMC100PPSerialCommunication, LoopSerialCommunication
):
    pass


class Pfa2FilterLoopSerialCommunication(
    Pfa2FilterSerialCommunication, LoopSerialCommunication
):
    pass


class PfeifferTPGLoopSerialCommunication(
    PfeifferTPGSerialCommunication, LoopSerialCommunication
):
    pass


class TechnixLoopSerialCommunication(
    TechnixSerialCommunication, LoopSerialCommunication
):
    pass


class KA3000PLoopCommunication(KA3000PCommunication, LoopSerialCommunication):
    pass



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
""" """

import logging

# from typing import Optional, cast
#
# from libtiepie import i2chost as ltp_i2c
#
# from .base import TiePieDeviceType, _require_dev_handle, wrap_libtiepie_exception
# from .oscilloscope import TiePieOscilloscope
# from .utils import PublicPropertiesReprMixin
#
logger = logging.getLogger(__name__)

msg = "The I2C-Host feature is (currently) not supported by the `python-libtiepie`"

logger.error(msg)

raise DeprecationWarning(msg)

#
#
# class TiePieI2CHostConfigLimits:
#     """
#     Default limits for I2C host parameters.
#     """
#
#     def __init__(self, dev_i2c: ltp_i2c.I2CHost) -> None:
#         # I2C Host
#         pass
#
#
# class TiePieI2CHostConfig(PublicPropertiesReprMixin):
#     """
#     I2C Host's configuration with cleaning of values in properties setters.
#     """
#
#     def __init__(self, dev_i2c: ltp_i2c.I2CHost):
#         self.dev_i2c: ltp_i2c.I2CHost = dev_i2c
#         self.param_lim: TiePieI2CHostConfigLimits = TiePieI2CHostConfigLimits(
#             dev_i2c=dev_i2c
#         )
#
#
# class TiePieI2CHostMixin:
#     """
#     TiePie I2CHost sub-device.
#
#     A wrapper for the `libtiepie.i2chost.I2CHost` class. To be mixed in with
#     `TiePieOscilloscope` base class.
#     """
#
#     def __init__(self, com, dev_config) -> None:
#         super().__init__(com, dev_config)  # type: ignore[call-arg]
#
#         self._i2c: Optional[ltp_i2c.I2CHost] = None
#
#         self.config_i2c: Optional[TiePieI2CHostConfig] = None
#         """
#         I2C host's dynamical configuration.
#         """
#
#     @_require_dev_handle(TiePieDeviceType.I2C)
#     def _i2c_config_setup(self) -> None:
#         """
#         Setup dynamical configuration for the connected I2C host.
#         """
#         self.config_i2c = TiePieI2CHostConfig(
#             dev_i2c=self._i2c,
#         )
#
#     def _i2c_config_teardown(self) -> None:
#         """
#         Teardown dynamical configuration for the I2C Host.
#         """
#         self.config_i2c = None
#
#     def _i2c_close(self) -> None:
#         if self._i2c is not None:
#             del self._i2c
#             self._i2c = None
#
#     def start(self) -> None:
#         """
#         Start the I2C Host.
#         """
#         super().start()
#         logger.info("Starting I2C host")
#
#         self._i2c = cast(TiePieOscilloscope, self)._get_device_by_serial_number(
#             TiePieDeviceType.I2C
#         )
#         self._i2c_config_setup()
#
#     @wrap_libtiepie_exception
#     def stop(self) -> None:
#         """
#         Stop the I2C host.
#         """
#         logger.info("Stopping I2C host")
#
#         self._i2c_config_teardown()
#         self._i2c_close()
#
#         super().stop()



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device class for Newport SMC100PP stepper motor controller with serial communication.

The SMC100PP is a single axis motion controller/driver for stepper motors up to 48 VDC
at 1.5 A rms. Up to 31 controllers can be networked through the internal RS-485
communication link.

Manufacturer homepage:
https://www.newport.com/f/smc100-single-axis-dc-or-stepper-motion-controller
"""

from .newport import (  # noqa: F401
    NewportConfigCommands,
    NewportControllerError,
    NewportMotorError,
    NewportMotorPowerSupplyWasCutError,
    NewportSerialCommunicationError,
    NewportSMC100PP,
    NewportSMC100PPConfig,
    NewportSMC100PPSerialCommunication,
    NewportSMC100PPSerialCommunicationConfig,
    NewportStates,
    NewportUncertainPositionError,
)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device class for controlling a MBW 973 SF6 Analyzer over a serial connection.

The MBW 973 is a gas analyzer designed for gas insulated switchgear and measures
humidity, SF6 purity and SO2 contamination in one go.
Manufacturer homepage: https://www.mbw.ch/products/sf6-gas-analysis/973-sf6-analyzer/
"""

from .mbw973 import (  # noqa: F401
    MBW973,
    MBW973Config,
    MBW973ControlRunningError,
    MBW973Error,
    MBW973PumpRunningError,
    MBW973SerialCommunication,
    MBW973SerialCommunicationConfig,
)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
SerialCommunication for FuG
"""

import logging
from typing import cast

from hvl_ccb.comm.base import SyncCommunicationProtocol
from hvl_ccb.comm.serial import (
    SerialCommunication,
    SerialCommunicationBytesize,
    SerialCommunicationConfig,
    SerialCommunicationParity,
    SerialCommunicationStopbits,
)
from hvl_ccb.configuration import configdataclass
from hvl_ccb.utils.typing import Number

from .errors import FuGError, FuGErrorcodes

logger = logging.getLogger(__name__)


@configdataclass
class FuGSerialCommunicationConfig(SerialCommunicationConfig):
    """
    Configuration dataclass for :class:`FuGSerialCommunication`.
    """

    #: Baudrate for FuG power supplies is 9600 baud
    baudrate: int = 9600

    #: FuG does not use parity
    parity: str | SerialCommunicationParity = SerialCommunicationParity.NONE

    #: FuG uses one stop bit
    stopbits: int | SerialCommunicationStopbits = SerialCommunicationStopbits.ONE

    #: One byte is eight bits long
    bytesize: int | SerialCommunicationBytesize = SerialCommunicationBytesize.EIGHTBITS

    #: The terminator is LF
    terminator: bytes = b"\n"

    #: use 3 seconds timeout as default
    timeout: Number = 3

    #: default time to wait between attempts of reading a non-empty text
    wait_sec_read_text_nonempty: Number = 0.5

    #: default number of attempts to read a non-empty text
    default_n_attempts_read_text_nonempty: int = 10


class FuGSerialCommunication(SerialCommunication, SyncCommunicationProtocol):
    """
    Specific communication protocol implementation for
    FuG power supplies.
    Already predefines device-specific protocol parameters in config.
    """

    @staticmethod
    def config_cls():
        return FuGSerialCommunicationConfig

    def query(
        self,
        command: str,
        n_attempts_max: int | None = None,
        attempt_interval_sec: Number | None = None,
    ) -> str | None:
        """
        Send a command to the interface and handle the status message.
        Raises an error, if the answer starts with "E".

        :param command: Command to send
        :raises FuGError: if the connection is broken or the error from the power
            source itself
        :return: Answer from the interface or empty string
        """

        with self.access_lock:
            logger.debug(f"FuG communication, send: {command}")
            answer: str | None = super().query(
                command, n_attempts_max, attempt_interval_sec
            )
            logger.debug(f"FuG communication, receive: {answer}")
            if not answer:
                cast("FuGErrorcodes", FuGErrorcodes.E504).raise_()
            try:
                FuGErrorcodes(cast("str", answer)).raise_()
            except ValueError as exc:
                logger.exception(
                    "ValueError at finding the correct FuGErrorcode", exc_info=exc
                )
                if cast("str", answer).startswith("E"):
                    msg = f'The unknown errorcode "{answer}" was detected.'
                    raise FuGError(msg) from exc
                return answer
            else:
                return ""



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
The very special "AdvancedCube" with its two single implementations "SynthCube" and
"FPCSCube".
"""

import logging
from collections.abc import Iterator

from hvl_ccb.configuration import configdataclass
from hvl_ccb.utils.enum import BoolEnum
from hvl_ccb.utils.typing import Number

from .base import BaseCube, BaseCubeConfiguration
from .constants import SafetyStatus, _Door
from .errors import (
    AdvancedCubeModuleError,
    ChargerSwitchError,
    DischargeError,
    ShortCircuitError,
)
from .switches import SwitchOperation, _Switch

logger = logging.getLogger(__name__)


class ActivationStatus(BoolEnum):
    """
    Boolean activation status of an advanced cube output bit
    """

    ACTIVATED = True
    DEACTIVATED = False


class PressurizationStatus(BoolEnum):
    """
    Boolean status of an advanced cube pressure sensor
    """

    PRESSURIZED = True
    DEPRESSURIZED = False


class _DischargeSwitch(_Switch):
    """
    Discharge switch within a module of the advanced cube.
    """

    _ERROR_CLS = DischargeError
    _STATES_FOR_SWITCHING = (
        SafetyStatus.RED_READY,
        SafetyStatus.RED_OPERATE,
    )

    def __init__(self, handle, module_number: int) -> None:
        super().__init__(handle, device_name=f"Discharge Switch {module_number}")
        self._CMD_STATUS = (
            f'"DB_Safety_Circuit"."Module_{module_number}".'
            '"Discharge_Circuit"."si_HMI_Status"'
        )
        self._CMD_OPERATING_STATUS = (
            f'"DB_Safety_Circuit"."Module_{module_number}"."sx_able_to_discharge"'
        )
        self._CMD_MANUAL = (
            f'"DB_Safety_Circuit"."Module_{module_number}"."sx_discharge"'
        )


class _ShortCircuitSwitch(_Switch):
    """
    Short-circuit switch within a module of the advanced cube.
    """

    _ERROR_CLS = ShortCircuitError

    def __init__(self, handle, module_number: int) -> None:
        identifier = (
            "Central_Earthing" if module_number == 0 else f"Module_{module_number}"
        )
        super().__init__(handle, device_name=f"Short-Circuit Switch {identifier}")
        identifier_extended = (
            identifier if module_number == 0 else identifier + '"."Short_Circuit'
        )
        self._CMD_STATUS = (
            f'"DB_Safety_Circuit"."{identifier_extended}"."si_HMI_Status"'
        )
        self._CMD_OPERATING_STATUS = (
            f'"DB_Safety_Circuit"."{identifier}"."sx_able_to_short_circuit"'
        )
        self._CMD_MANUAL = None


class _ChargerSwitch(_Switch):
    """
    Charger switch within a module of the advanced cube.
    """

    _ERROR_CLS = ChargerSwitchError
    _STATES_FOR_SWITCHING = (SafetyStatus.RED_OPERATE,)

    def __init__(self, handle, module_number: int) -> None:
        identifier = (
            "Central_Charger" if module_number == 0 else f"Module_{module_number}"
        )
        super().__init__(handle, device_name=f"Charger Switch {identifier}")
        identifier_extended = (
            identifier if module_number == 0 else identifier + '"."Charger_Circuit'
        )
        self._CMD_STATUS = (
            f'"DB_Safety_Circuit"."{identifier_extended}"."si_HMI_Status"'
        )
        self._CMD_OPERATING_STATUS = (
            f'"DB_Safety_Circuit"."{identifier}"."sx_able_to_charge"'
        )
        self._CMD_MANUAL = f'"DB_Safety_Circuit"."{identifier}"."sx_charge"'


class _Module:
    """
    Module level interface between AdvancedCube and current sources.
    """

    _ACTIVATABLE_STATES = (
        SafetyStatus.GREEN_NOT_READY,
        SafetyStatus.GREEN_READY,
    )

    def __init__(self, handle, index: int) -> None:
        self._handle: AdvancedCube = handle
        self._index: int = index
        self.discharge_switch: _DischargeSwitch = _DischargeSwitch(handle, index)
        self.short_circuit_switch: _ShortCircuitSwitch = _ShortCircuitSwitch(
            handle, index
        )
        self.charger_switch: _ChargerSwitch = _ChargerSwitch(handle, index)
        self._CMD_ACTIVE: str = f'"DB_Safety_Circuit"."Module_{index}"."sx_dial"'
        self._CMD_VOLTAGE: str = f'"DB_Safety_Circuit"."Module_{index}"."si_voltage"'
        self._CMD_MAX_SC_SWITCHING_VOLTAGE: str = (
            f'"DB_Safety_Circuit"."Module_{index}".'
            '"Ir_max_short_circuit_switching_voltage"'
        )

    def __repr__(self) -> str:
        return f"Module {self._index}"

    @property
    def active(self) -> ActivationStatus:
        """
        Module activation status

        :return: boolean module activation status
        """
        activation_status = ActivationStatus(self._handle.read(self._CMD_ACTIVE))
        logger.info(f"Module {self._index} is {activation_status.name}")
        return activation_status

    @active.setter
    def active(self, new_activation_status: ActivationStatus):
        new_activation_status = ActivationStatus(new_activation_status)
        if self._handle.status in self._ACTIVATABLE_STATES:
            self._handle.write(self._CMD_ACTIVE, new_activation_status)
            logger.info(f"Module {self._index} set to {new_activation_status.name}")
        else:
            msg = (
                f"Tried to set Module {self._index} to {new_activation_status.name}. "
                "Safety circuit must be in green state to change the activation "
                f"status of a module, but is in {self._handle.status.name}"
            )
            logger.error(msg)
            raise AdvancedCubeModuleError(msg)

    @property
    def voltage(self) -> int:
        """
        Measured voltage of a module.

        :return: measured voltage in V
        """
        voltage = int(self._handle.read(self._CMD_VOLTAGE))
        logger.info(f"Module {self._index} voltage: {voltage} V")
        return voltage

    @property
    def max_short_circuit_switching_voltage(self) -> int:
        """
        Maximum switching voltage of a module.

        :return: maximum switching voltage in V
        """
        voltage = int(self._handle.read(self._CMD_MAX_SC_SWITCHING_VOLTAGE))
        logger.info(
            f"Maximum short-circuit switching voltage of Module {self._index}"
            f" is set to {voltage} V"
        )
        return voltage


class _Pneumatics:
    """
    Pneumatic circuit of a current source controlled by an AdvancedCube
    """

    _CMD_STATUS = '"Ix_Allg_pressure_ok"'
    _CMD_ACTIVE = '"Qx_Allg_Main_Pressure"'

    def __init__(self, handle) -> None:
        self._handle: AdvancedCube = handle

    @property
    def status(self) -> PressurizationStatus:
        """
        Pneumatic circuit pressure status

        :return: boolean pressurization status
        """
        status = PressurizationStatus(self._handle.read(self._CMD_STATUS))
        logger.info(f"Pneumatic circuit is {status.name}")
        return status

    @property
    def active(self) -> ActivationStatus:
        """
        Pressure valve activation status

        :return: boolean pressure valve activation status
        """
        activation_status = ActivationStatus(self._handle.read(self._CMD_ACTIVE))
        logger.info(f"Main pressure valve is {activation_status.name}")
        return activation_status

    @active.setter
    def active(self, new_activation_status: ActivationStatus):
        new_activation_status = ActivationStatus(new_activation_status)
        self._handle.write(self._CMD_ACTIVE, new_activation_status)
        logger.info(f"Main pressure valve {new_activation_status.name}")


class _Siren:
    """
    Acoustic alert device controlled by an AdvancedCube
    """

    def __init__(self, handle, index: int, name: str) -> None:
        self._handle: AdvancedCube = handle
        self._name = name
        self._CMD_ACTIVE: str = f'"Qx_Allg_Acoustic_{index}"'

    @property
    def active(self) -> ActivationStatus:
        """
        Siren activation status

        :return: boolean siren activation status
        """
        active = ActivationStatus(self._handle.read(self._CMD_ACTIVE))
        logger.info(f"{self._name} siren is {active.name}")
        return active

    @active.setter
    def active(self, active: ActivationStatus):
        active = ActivationStatus(active)
        self._handle.write(self._CMD_ACTIVE, active)
        logger.info(f"{self._name} siren has been {active.name}")


@configdataclass
class AdvancedCubeConfiguration(BaseCubeConfiguration):
    """Adaption of base configuration to accept longer time to wait for a discharge"""

    timeout_status_change: Number = 20


class AdvancedCube(BaseCube):
    """
    The AdvancedCube is the template for its two special implementations
    """

    _N_MODULES: int | None = None

    def __init__(self, com, dev_config=None) -> None:
        super().__init__(com, dev_config)
        self.modules: dict[int, _Module] = {}
        self.pneumatics: _Pneumatics | None = None
        self.charging_siren: _Siren | None = None
        self.triggering_siren: _Siren | None = None

    @staticmethod
    def config_cls():
        return AdvancedCubeConfiguration

    @property
    def number_of_modules(self) -> int:
        """
        Queries the number of installed modules (active plus inactive)
        """
        if self._N_MODULES is None:
            msg = "No number of modules was set in implementation."
            logger.error(msg)
            raise AdvancedCubeModuleError(msg)

        val = self._N_MODULES
        logger.info(f"{self.__class__.__name__} has {val} modules installed.")
        return val

    def start(self) -> None:
        """
        Start the connection to the AdvancedCube and set up the modules
        """
        super().start()

        try:
            _ = self.number_of_modules
        except AdvancedCubeModuleError:
            self.stop()
            raise

        for ii in range(self.number_of_modules):
            self.modules[ii + 1] = _Module(self, ii + 1)

        self.pneumatics = _Pneumatics(self)
        self.charging_siren = _Siren(self, 1, name="Charging")
        self.triggering_siren = _Siren(self, 2, name="Triggering")

    @property
    def active_modules(self) -> Iterator[_Module]:
        """
        Iterator for all active modules
        """
        for module in self.modules.values():
            if module.active == ActivationStatus.ACTIVATED:
                yield module

    @property
    def operate_discharge_switches(self) -> list[SwitchOperation]:
        """
        Property to query all discharge switches of the installed/active modules

        :return: `list[SwitchOperation]` of the individual discharge switches
        """
        return [module.discharge_switch.operate for module in self.active_modules]

    @operate_discharge_switches.setter
    def operate_discharge_switches(self, operation: SwitchOperation) -> None:
        """
        Operate all discharge switches of the installed/active modules

        :param operation: a `SwitchOperation` to `OPEN` or `CLOSE`
        """
        for module in self.active_modules:
            module.discharge_switch.operate = operation


class SynthCube(AdvancedCube):
    """
    The SynthCube is the special Cube for the Synthetic Circuit
    """

    _N_MODULES = 4

    def __init__(self, com, dev_config=None) -> None:
        super().__init__(com, dev_config)
        self.hcc_loop_1: _Module | None = None
        self.hcc_loop_2: _Module | None = None
        self.hcc_loop_3: _Module | None = None
        self.cic: _Module | None = None

    def start(self) -> None:
        super().start()
        self.hcc_loop_1 = self.modules[1]
        self.hcc_loop_2 = self.modules[2]
        self.hcc_loop_3 = self.modules[3]
        self.cic = self.modules[4]

    @property
    def active_hcc_loops(self) -> Iterator[_Module]:
        """
        Iterator for all active high-current modules
        """
        for module in self.active_modules:
            if module in (self.hcc_loop_1, self.hcc_loop_2, self.hcc_loop_3):
                yield module


class FPCSCube(AdvancedCube):
    """
    The FPCSCube is the special Cube for the FPCS
    """

    _N_MODULES = 8

    door_fpcs_status = _Door("FPCS", "door status")

    def __init__(self, com, dev_config=None) -> None:
        super().__init__(com, dev_config)
        self.central_earthing: _ShortCircuitSwitch | None = None
        self.central_charger: _ChargerSwitch | None = None
        self.module: dict[str, _Module] = {}

    def start(self) -> None:
        super().start()
        for ii in range(self._N_MODULES):
            self.module[chr(ord("A") + ii)] = self.modules[ii + 1]
            self.module[chr(ord("a") + ii)] = self.modules[ii + 1]
        self.central_charger = _ChargerSwitch(self, 0)
        self.central_earthing = _ShortCircuitSwitch(self, 0)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Module with base classes for communication protocols.
"""

import logging
from abc import ABC, abstractmethod
from threading import RLock
from time import sleep

from typing_extensions import Self

from hvl_ccb.configuration import ConfigurationMixin, EmptyConfig, configdataclass
from hvl_ccb.error import CCBError
from hvl_ccb.utils.typing import Number

logger = logging.getLogger(__name__)


class CommunicationError(CCBError):
    pass


class CommunicationProtocol(ConfigurationMixin, ABC):
    """
    Communication protocol abstract base class.

    Specifies the methods to implement for communication protocol, as well as
    implements some default settings and checks.
    """

    def __init__(self, config) -> None:
        """
        Constructor for CommunicationProtocol. Takes a configuration dict or
        configdataclass as the single parameter.

        :param config: Configdataclass or dictionary to be used with the default
            config dataclass.
        """

        super().__init__(config)

        #: Access lock to use with context manager when
        #: accessing the communication protocol (thread safety)
        self.access_lock = RLock()

    # TECH: to be uncommented; pending for v1.0
    # @property
    # @abstractmethod
    # def is_open(self) -> Optional[bool]:
    #     pass  # pragma: no cover

    @abstractmethod
    def open(self) -> None:
        """
        Open communication protocol
        """

    @abstractmethod
    def close(self) -> None:
        """
        Close the communication protocol
        """

    def __enter__(self) -> Self:
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()


class NullCommunicationProtocol(CommunicationProtocol):
    """
    Communication protocol that does nothing.
    """

    def open(self) -> None:
        """
        Void open function.
        """

    def close(self) -> None:
        """
        Void close function.
        """

    @staticmethod
    def config_cls() -> type[EmptyConfig]:
        """
        Empty configuration

        :return: EmptyConfig
        """
        return EmptyConfig


@configdataclass
class AsyncCommunicationProtocolConfig:
    """
    Base configuration data class for asynchronous communication protocols
    """

    #: The terminator character. Typically this is ``b'\r\n'`` or ``b'\n'``, but can
    #: also be ``b'\r'`` or other combinations. This defines the end of a single line.
    terminator: bytes = b"\r\n"

    #: Standard encoding of the connection. Typically this is ``utf-8``, but can also
    #: be ``latin-1`` or something from here:
    #: https://docs.python.org/3/library/codecs.html#standard-encodings
    encoding: str = "utf-8"
    #: Encoding error handling scheme as defined here:
    #: https://docs.python.org/3/library/codecs.html#error-handlers
    #: By default strict error handling that raises `UnicodeError`.
    encoding_error_handling: str = "strict"

    #: time to wait between attempts of reading a non-empty text
    wait_sec_read_text_nonempty: Number = 0.5

    #: default number of attempts to read a non-empty text
    default_n_attempts_read_text_nonempty: int = 10

    def clean_values(self) -> None:
        try:
            "".encode(encoding=self.encoding)
        except LookupError as exc:
            raise ValueError(str(exc)) from exc

        try:
            "\ufffd".encode(encoding="ascii", errors=self.encoding_error_handling)
        except LookupError as exc:
            raise ValueError(str(exc)) from exc
        except UnicodeEncodeError:
            # expected error
            pass

        if self.wait_sec_read_text_nonempty <= 0:
            msg = (
                "Wait time between attempts to read a non-empty text must be be a "
                "positive value (in seconds)."
            )
            raise ValueError(msg)

        if self.default_n_attempts_read_text_nonempty <= 0:
            msg = (
                "Default number of attempts of reading a non-empty text must be a "
                "positive integer."
            )
            raise ValueError(msg)


class AsyncCommunicationProtocol(CommunicationProtocol):
    """
    Abstract base class for asynchronous communication protocols
    """

    @staticmethod
    def config_cls() -> type[AsyncCommunicationProtocolConfig]:
        return AsyncCommunicationProtocolConfig

    @abstractmethod
    def read_bytes(self) -> bytes:
        """
        Read a single line as `bytes` from the communication.

        This method uses `self.access_lock` to ensure thread-safety.

        :return: a single line as `bytes` containing the terminator, which can also be
            empty b""
        """

    @abstractmethod
    def write_bytes(self, data: bytes) -> int:
        """
        Write data as `bytes` to the communication.

        This method uses `self.access_lock` to ensure thread-safety.

        :param data: data as `bytes`-string to be written
        :return: number of bytes written
        """

    def read(self) -> str:
        """
        Read a single line of text as `str` from the communication.

        :return: text as `str` including the terminator, which can also be empty ""
        """
        return self.read_bytes().decode(
            encoding=self.config.encoding, errors=self.config.encoding_error_handling
        )

    def write(self, text: str) -> None:
        """
        Write text as `str` to the communication.

        :param text: test as a `str` to be written
        """
        self.write_bytes(
            text.encode(
                encoding=self.config.encoding,
                errors=self.config.encoding_error_handling,
            )
            + self.config.terminator
        )

    def read_nonempty(
        self,
        n_attempts_max: int | None = None,
        attempt_interval_sec: Number | None = None,
    ) -> str | None:
        """
        Try to read a non-empty single line of text as `str` from the communication.
        If the host does not reply or reply with white space only, it will return None.

        :return: a non-empty text as a `str` or `None` in case of an empty string
        :param n_attempts_max: Amount of attempts how often a non-empty text is tried to
            be read
        :param attempt_interval_sec: time between the reading attempts
        """
        if n_attempts_max is None:
            n_attempts_max = self.config.default_n_attempts_read_text_nonempty
        if attempt_interval_sec is None:
            attempt_interval_sec = self.config.wait_sec_read_text_nonempty

        answer = self.read().strip()

        while len(answer) == 0 and n_attempts_max > 0:
            sleep(attempt_interval_sec)
            answer = self.read().strip()
            n_attempts_max -= 1
        if answer == "":
            return None  # Return None for an empty String
        return answer

    def read_all(
        self,
        n_attempts_max: int | None = None,
        attempt_interval_sec: Number | None = None,
    ) -> str | None:
        """
        Read all lines of text from the connection till nothing is left to read.

        :param n_attempts_max: Amount of attempts how often a non-empty text is tried to
            be read
        :param attempt_interval_sec: time between the reading attempts
        :return: A multi-line `str` including the terminator internally
        """

        answer = self.read_nonempty(
            n_attempts_max=n_attempts_max,
            attempt_interval_sec=attempt_interval_sec,
        )
        result = ""
        newline = self.config.terminator.decode(
            encoding=self.config.encoding, errors=self.config.encoding_error_handling
        )
        while answer:
            result = f"{result}{newline}{answer}"
            answer = self.read_nonempty(
                n_attempts_max=n_attempts_max,
                attempt_interval_sec=attempt_interval_sec,
            )

        if result == "":
            return None
        return f"{result.strip()}{self.config.terminator}"

    def read_text(self) -> str:
        """
        Read one line of text from the serial port. The input buffer may
        hold additional data afterwards, since only one line is read.

        NOTE: backward-compatibility proxy for `read` method; to be removed in v1.0

        :return: String read from the serial port; `''` if there was nothing to read.
        :raises SerialCommunicationIOError: when communication port is not opened
        """
        return self.read()

    def write_text(self, text: str) -> None:
        """
        Write text to the serial port. The text is encoded and terminated by
        the configured terminator.

        NOTE: backward-compatibility proxy for `read` method; to be removed in v1.0

        :param text: Text to send to the port.
        :raises SerialCommunicationIOError: when communication port is not opened
        """
        return self.write(text)

    def read_text_nonempty(
        self,
        n_attempts_max: int | None = None,
        attempt_interval_sec: Number | None = None,
    ) -> str | None:
        """
        Reads from the serial port, until a non-empty line is found, or the number of
        attempts is exceeded.

        NOTE: backward-compatibility proxy for `read` method; to be removed in v1.0

        Attention: in contrast to `read_text`, the returned answer will be stripped of
        a whitespace newline terminator at the end, if such terminator is set in
        the initial configuration (default).

        :param n_attempts_max: maximum number of read attempts
        :param attempt_interval_sec: time between the reading attempts
        :return: String read from the serial port; `''` if number of attempts is
            exceeded or serial port is not opened.
        """
        return self.read_nonempty(
            n_attempts_max=n_attempts_max,
            attempt_interval_sec=attempt_interval_sec,
        )


class SyncCommunicationProtocolConfig(AsyncCommunicationProtocolConfig):
    """
    Base configuration data class for synchronous communication protocols
    """


class SyncCommunicationProtocol(AsyncCommunicationProtocol, ABC):
    """
    Abstract base class for synchronous communication protocols with `query()`
    """

    @staticmethod
    def config_cls() -> type[SyncCommunicationProtocolConfig]:
        return SyncCommunicationProtocolConfig

    def query(
        self,
        command: str,
        n_attempts_max: int | None = None,
        attempt_interval_sec: Number | None = None,
    ) -> str | None:
        """
        Send a command to the interface and handle the status message.
        Possibly raises an exception.

        :param command: Command to send
        :param n_attempts_max: Amount of attempts how often a non-empty text is tried to
            be read as answer
        :param attempt_interval_sec: time between the reading attempts
        :return: Answer from the interface, which can be None instead of an empty reply
        """

        with self.access_lock:
            self.write(text=command)
            answer: str | None = self.read_nonempty(
                n_attempts_max=n_attempts_max, attempt_interval_sec=attempt_interval_sec
            )
            # expects an answer string or None

            return answer



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Mock OscilloscopeChannelTrigger
"""

from libtiepie import const
from libtiepie.oscilloscopechanneltrigger import (
    OscilloscopeChannelTrigger as LtpOscilloscopeChannelTrigger,
)

from .oscilloscopechanneltriggerhystereses import OscilloscopeChannelTriggerHystereses
from .oscilloscopechanneltriggerlevels import OscilloscopeChannelTriggerLevels


class OscilloscopeChannelTrigger(LtpOscilloscopeChannelTrigger):
    def __init__(self) -> None:
        self._enabled = False
        self._hystereses = OscilloscopeChannelTriggerHystereses()
        self._levels = OscilloscopeChannelTriggerLevels()
        self._kind = const.TK_ANYEDGE
        self._level_mode = const.TLM_UNKNOWN

    def _get_enabled(self):
        return self._enabled

    def _set_enabled(self, value):
        self._enabled = value

    def _get_hystereses(self):
        return self._hystereses

    def _get_levels(self):
        return self._levels

    def _get_kind(self):
        return self._kind

    def _set_kind(self, value):
        self._kind = value

    def _get_level_mode(self):
        return self._level_mode

    def _set_level_mode(self, value):
        self._level_mode = value

    enabled = property(_get_enabled, _set_enabled)
    hystereses = property(_get_hystereses)
    levels = property(_get_levels)
    kind = property(_get_kind, _set_kind)
    level_mode = property(_get_level_mode, _set_level_mode)
